// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*
 Epspatialcti

   Authors:  Richard Saxton - XMM SOC
         
   Description: Corrects an EPIC-pn event file for spatially-dependent 
                CTI effects
   Comments: 
 
  $Id: Epspatialcti.cc,v 1.2 2012/03/22 10:13:02 rsaxton Exp $
*/

#include "CalServer.h"
#include "Epspatialcti.h"

#include <Tracer.h>
#include <errstr.h>
#include <param.h>

Epspatialcti::Epspatialcti()
{
PROC

   /* Find if the Sanders, Denerl et al. 2020 method is wanted */
    bool newmode = booleanParameter("SD20mode");

  /* Get the name of the input and output files */
    std::string inSetName = stringParameter("table");
    std::string outSetName = stringParameter("outset");

  /* Copy the input to the output file */
    errstr<<"Copying data ... "<<Sas::message(Msg::App, Msg::Verbose);

    DataSet *set = dataSetServer->clone(inSetName, outSetName, Dal::Modify);

   /* Access the events table and read columns */
    Table *tb = set->table("EVENTS");

   /* Check if this file has already been corrected for spatial CTI */
    if (tb->hasAttribute("SPATCTIC"))
         errstr<<"Event file "<<stringParameter("table")<<" has already "<<
                 "been corrected for spatial CTI effects - file will not "<<
                 "be changed" << Sas::error("AlreadyCorrected");
 
    Column *colrx = tb->column("RAWX");
    ColumnData *crawx= colrx->data();
    int16 *rawx = crawx->int16Data();
    Column *colry = tb->column("RAWY");
    ColumnData *crawy= colry->data();
    int16 *rawy = crawy->int16Data();
    Column *colccd = tb->column("CCDNR");
    ColumnData *cccd= colccd->data();
    int8    *ccdId  = cccd->int8Data();
    Column *colpi = tb->column("PI");
    ColumnData *cpi= colpi->data();
    int16    *pi  = cpi->int16Data();

    int N = tb->rows();

   /* Get the observation mode from the event file */
    string submode = "PrimeFullWindow";
    if (set->hasAttribute("SUBMODE"))
      {
         submode = set->attribute("SUBMODE")->asString();
      } else {
         submode = " ";
      }

    Xmm::Mode windowMode = Xmm::toMode(submode);   // Exits if submode invalid

   /* Set the CAL */
//    Xmm::Instrument instrument=Xmm::EPN;
    calServer.setInstrument(Xmm::EPN);
    calServer.state()->mode()->set(windowMode);

   /* Get the gain correction atom */
    EnergyCorrector *gainCorrector = 0;
    gainCorrector = calServer.getAtom(gainCorrector);

   /* Set observation date and get data from correct element
       if new mode required */
    if (newmode) {
    
      /* Set the observation date */
      if (set->hasAttribute("REVOLUT"))
         calServer.state()->exposureRevolution()->set(
                     set->attribute("REVOLUT")->asInt());
      else 
         errstr<<"Event file does not contain keyword REVOLUT which "
               <<"is needed for the time-dependent spatial correction"
               << Sas::error("NoRevolutionKey");

      string element = stringParameter("element");
      gainCorrector->spatialSD20SetElement(element);

      /* Mn can not be used with extended Full Frame mode */
      if (element == "MN" && windowMode == Xmm::PRIME_FULL_WINDOW_EXTENDED)
         errstr<<"The Manganese calibration is not available for " 
               <<"Extended Full Frame mode. Try using the Cu offsets" 
               << Sas::error("IncompatibleElementMode");
    }

   /* Debug */
    //Ccd ccd = ccdId[0];
    //cout<<"IN: "<<rawx[3]<<" "<<rawy[3]<<" "<<ccd<<" "<<pi[3]<<endl;
    
   /* Loop over CCDs */
    //unsigned nCCDs  = instrument.nCcds();
    vector<int> elist;
    const XmmInstrument *instrument = calServer.instrument();
    unsigned nCCDs  = instrument->nCcds();

    for (Ccd ccd=1; ccd<=nCCDs; ++ccd) {  // Change to get from CAL

      /* Find the events which fall on this CCD */
        elist.clear();

        for (int n=0; n<N; ++n) 
          if (ccdId[n] == INT(ccd)) 
              elist.push_back(n);

      /* Set up a vector to hold the PI events as a float vector */
        CalReal32Vector rpiv(0, elist.size());
        const CalInt16Vector xv(0, elist.size());
        const CalInt16Vector yv(0, elist.size());


        for (unsigned e=0; e<elist.size(); ++e) {
             rpiv[e] = float(pi[elist[e]]);
             xv[e] =  rawx[elist[e]];
             yv[e] =  rawy[elist[e]];
        }

    //    cout<<"CCD: "<<ccd<<" "<<" [1]: "<<elist[1]<<" "<<pi[elist[1]]<<endl;
    //    cout<<"CCD: "<<ccd<<" "<<" size: "<<elist.size()<<" "<<rpiv[1]<<endl;

      /* Set this CCD */
        calServer.state()->ccd()->set(ccd);

         //cout<<"LAST: "<<ccd<<" "<<xv[elist.size()-1]<<" "<<yv[elist.size()-1]<<" "<<rpiv[elist.size()-1]<<endl;

      /* Make the spatial CTI correction */
        if (newmode) 
           gainCorrector->spatialCtiCorrectSD20(rpiv, xv, yv);
        else
           gainCorrector->spatialCtiCorrect(rpiv, xv, yv);

      /* Set the energies back to PIs */
        for (unsigned e=0; e<elist.size(); ++e) 
            pi[elist[e]] = floor(rpiv[e]+0.5);
    } // End of loop over ccds

   /* If everything has worked, add a keyword to the file header to 
      say that it has been corrected */
    *tb->addAttribute("SPATCTIC",  
                      "Has the spatial CTI correction been applied?") = "TRUE";

   /* Clean up */
    calServer.releaseAtom(gainCorrector);

    delete crawx;
    delete crawy;
    delete cccd;
    delete cpi;

    dataSetServer->close(set);

ENDPROC
}

/* Destructor */
Epspatialcti:: ~Epspatialcti()
{
  PROC;

  ENDPROC;
}

